-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
-- Soubor:  pil032.hs                                                        --
-- Nazev:   Prevod nedeterministickeho konecneho automatu na deterministicky --
-- Autor:   David Pilar                                                      --
-- Login:   pil032                                                           --
-- Skupina: L394                                                             --
-- Datum:   6.4.2003                                                         --
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------

module Pil032 where

import IO

--------------------
-- Testovaci data --
--------------------

ndka1 = [([-1,1],[1,2],[1]),([2],[3],[]),([3],[],[4]),([-2,4],[],[]),([-1,5],[5],[5,6]),([6],[],[7]),([7],[],[8]),([8],[],[9]),([-2,9],[9],[9])]
ndka2 = [([-1,1],[1,2],[1]),([2],[3],[1,2]),([3],[],[4]),([-2,4],[],[5]),([-1,5],[5,2],[5])]
ndka3 = [([-1,1],[2],[1]),([-2,2],[5],[5]),([-2,3],[5],[5]),([-1,4],[4],[3]),([5],[],[])]

--------------------
-- Hlavni program --
--------------------

type Aut a = [([a],[a],[a])]

-----------------------------------------------------
-- Funkce pro vytvoreni deterministickeho automatu --
-----------------------------------------------------

-- nalezeni pocatecnich stavu --
pocatecni :: Aut Integer -> [Integer]
pocatecni [] = []
pocatecni (((x:xs),_,_):ys)
	| x == -1 = xs ++ pocatecni ys
	| otherwise = pocatecni ys

-- nalezeni prijimacich stavu --
prijimaci :: Aut Integer -> [Integer]
prijimaci [] = []
prijimaci (((x:xs),_,_):ys)
	| x == -2 = xs ++ prijimaci ys
	| otherwise = prijimaci ys

-- vymazani oznaceni pocatecnich stavu --
vymaz_pocatecni :: Aut Integer -> Aut Integer
vymaz_pocatecni [] = []
vymaz_pocatecni (((x:xs),ys,zs):us)
	| x == -1 = [(xs,ys,zs)] ++ vymaz_pocatecni us
	| otherwise = [((x:xs),ys,zs)] ++ vymaz_pocatecni us

-- vymazani oznaceni prijimacich stavu -- 
vymaz_prijimaci :: Aut Integer -> Aut Integer
vymaz_prijimaci [] = []
vymaz_prijimaci (((x:xs),ys,zs):us)
	| x == -2 = [(xs,ys,zs)] ++ vymaz_prijimaci us
	| otherwise = [((x:xs),ys,zs)] ++ vymaz_prijimaci us

-- vymazani pocatecnich a prijimacich stavu --
vymaz :: Aut Integer -> Aut Integer
vymaz xs = (vymaz_prijimaci . vymaz_pocatecni) xs

-- oznaceni prijimacich stavu --
oznac_prijimaci :: Aut Integer -> [Integer] -> Aut Integer
oznac_prijimaci [] _ = []
oznac_prijimaci ((xs,ys,zs):us) vs
	| porov vs xs = [(([-2] ++ xs),ys,zs)] ++ oznac_prijimaci us vs
	| otherwise = [(xs,ys,zs)] ++ oznac_prijimaci us vs

porov :: [Integer] -> [Integer] -> Bool
porov [] _ = False
porov (x:xs) ys
	| elem x ys = True
	| otherwise = porov xs ys

-- oznaceni pocatecniho stavu --
oznac_pocatecni :: Aut Integer -> Aut Integer
oznac_pocatecni ((xs,ys,zs):us) = (([-1] ++ xs),ys,zs):us

-- oznaceni pocatecniho a prijimacich stavu --
oznac :: Aut Integer -> [Integer] -> Aut Integer
oznac xs ys = oznac_pocatecni (oznac_prijimaci xs ys)

-- vytvoreni stavu pro prechod pres a --
stav0 :: Aut Integer -> [Integer] -> [Integer]
stav0 xs ys = opak (stav0_opak xs ys)

stav0_opak :: Aut Integer -> [Integer] -> [Integer]
stav0_opak [] _ = []
stav0_opak (((x:xs),ys,zs):us) vs
	| elem x vs = ys ++ stav0_opak us vs
	| otherwise = stav0_opak us vs

-- vytvoreni stavu pro prechod pres b --
stav1 :: Aut Integer -> [Integer] -> [Integer]
stav1 xs ys = opak (stav1_opak xs ys)

stav1_opak :: Aut Integer -> [Integer] -> [Integer]
stav1_opak [] _ = []
stav1_opak (((x:xs),ys,zs):us) vs
	| elem x vs = zs ++ stav1_opak us vs
	| otherwise = stav1_opak us vs
	
-- odstraneni opakujicich se prvku v seznamu --
opak :: [Integer] -> [Integer]
opak [] = []
opak (x:xs)
	| notElem x xs = [x] ++ opak xs
	| otherwise = opak xs

-- vytvoreni automatu --
vytvoreni :: Aut Integer -> [[Integer]] -> [[Integer]] -> Aut Integer
vytvoreni xs ys [] = []
vytvoreni xs ys (z:zs)
	= [(z, (stav0 xs z), (stav1 xs z))] 
	++ vytvoreni xs (vlozeni_vsech ys (stav0 xs z) (stav1 xs z)) 
	(vlozeni_vybir ys zs (stav0 xs z) (stav1 xs z)) 
	--(zs ++ [(stav0 xs z)] ++ [(stav1 xs z)]) 
	
-- vlozeni vstupniho stavu do seznamu vsech zaznamenavanych vstupu --
vlozeni_vsech :: [[Integer]] -> [Integer] -> [Integer] -> [[Integer]]
vlozeni_vsech xs ys zs = vlozeni2 (vlozeni2 xs ys) zs

vlozeni2 :: [[Integer]] -> [Integer] -> [[Integer]]
vlozeni2 xs ys
	| (porovn xs ys) = xs
	| otherwise = xs ++ [ys]

porovn :: [[Integer]] -> [Integer] -> Bool
porovn [] _ = False
porovn (x:xs) ys
	| (x == ys) = True
	| otherwise = porovn xs ys

-- vlozeni vstupnich stavu do seznamu vybiranych vstupu --
vlozeni_vybir :: [[Integer]] -> [[Integer]] -> [Integer] -> [Integer] -> [[Integer]]
vlozeni_vybir xs ys zs us = vlozeni3 xs (vlozeni3 xs ys zs) us

vlozeni3 :: [[Integer]] -> [[Integer]] -> [Integer] -> [[Integer]]
vlozeni3 xs ys zs
	| (porovn xs zs) = ys
	| otherwise = ys ++ [zs]
	
-- odstraneni stejnych stavu --
del :: Aut Integer -> Aut Integer
del [] = []
del (x:xs)
	| (srovnani x xs) = del xs
	| otherwise = [x] ++ (del xs)

srovnani :: ([Integer],[Integer],[Integer]) -> Aut Integer -> Bool
srovnani _ [] = False
srovnani xs (y:ys)
	| (xs == y) = True
	| otherwise = srovnani xs ys
	
-- prevod dka do normovaneho tvaru (pouze preznaceni) --
-- automat se prevodem z ndka do dka jiz v normovanem tvaru nachazi --
normovany :: Aut Integer -> Aut Integer
normovany xs = vloz (zapamatuj xs) (norma (vloz_cisla (odstran xs) (spoj (odstran xs))) 1)

-- vlozi cisla stavu normovaneho automatu, na ktere prislusne stavy ukazuji --
vloz_cisla :: Aut Integer -> 
	[([Integer],[Integer],[Integer],[Integer],[Integer])] ->
	[([Integer],[Integer],[Integer],[Integer],[Integer])]
vloz_cisla ns [] = []
vloz_cisla ns ((xs,us,ys,vs,zs):ws) = 
	[(xs,[(najdi ns ys 1)],ys,[(najdi ns zs 1)],zs)] ++ vloz_cisla ns ws

-- najde cislo stavu normovaneho automatu --
najdi :: Aut Integer -> 
	[Integer] -> Integer -> Integer
najdi ((xs,ys,zs):ws) ts c
	| (xs == ts) = c
	| otherwise = najdi ws ts (c+1)

-----------------------------------------------
-- Funkce pro vytvoreni normovaneho automatu --
-----------------------------------------------
	
-- vlozi prazdne seznamy, do kterych se budou zapisovat cisla stavu
-- normovaneho automatu
spoj :: Aut Integer -> [([Integer],[Integer],[Integer],[Integer],[Integer])]
spoj [] = []
spoj ((xs,ys,zs):us) = [(xs,[],ys,[],zs)] ++ spoj us

-- vytvori normovany automat (bez oznaceni pocatecnich a koncovych stavu) --
norma :: [([Integer],[Integer],[Integer],[Integer],[Integer])] ->
	Integer -> Aut Integer
norma [] _ = []
norma ((xs,us,ys,vs,zs):ws) c = [([c],us,vs)] ++ norma ws (c+1)

-- zapamatuje se pocatecni a koncove stavy --
zapamatuj :: Aut Integer -> [[Integer]]
zapamatuj [] = []
zapamatuj ((xs,ys,zs):ws) = [filter (<0) xs] ++ zapamatuj ws

-- odstrani oznaceni koncovych a pocatecnich stavu --
odstran :: Aut Integer -> Aut Integer
odstran [] = []
odstran ((xs,ys,zs):ws) = 
	[((filter (>0) xs),(filter (>0) ys),(filter (>0) zs))]
	++ odstran ws
	
-- vlozi oznaceni pocatecnich a koncovych stavu --
vloz :: [[Integer]] -> Aut Integer -> Aut Integer
vloz [] _ = []
vloz (u:us) ((xs,ys,zs):ws) = [((u ++ xs),ys,zs)] ++ vloz us ws

-----------------------------------------------------------
-- Funkce pro spusteni programu pro vytvoreni DKA z NDKA --
-----------------------------------------------------------

-- hlavni funkce, vytvori z NDKA DKA --
main :: Aut Integer -> Aut Integer
main xs = del (oznac (vytvoreni (vymaz xs) 
	[(pocatecni xs)] [(pocatecni xs)]) (prijimaci xs))
	
-- hlavni funkce, vytvori z NDKA DKA a prevede jej do normovaneho tvaru --
main2 :: Aut Integer -> Aut Integer
main2 xs = normovany (del (oznac (vytvoreni (vymaz xs) 
	[(pocatecni xs)] [(pocatecni xs)]) (prijimaci xs)))

-- hlavni funkce, vytvori z NDKA DKA, prevede jej do normovaneho tvaru --
-- a vypise zadany automat, automat DKA a automat v normovanem tvaru --
main3 :: Aut Integer -> IO()
main3 xs = do putStr ("\nLEGENDA:\n\n" ++
		"=== stav automatu === prechod na a === prechod na b ===\n\n" ++
		"Puvodni automat:\n\n" ++
		(aut2str xs) ++
		"\n\nPrevedeny na deterministicky:\n\n" ++
		(aut2str (del (oznac (vytvoreni (vymaz xs) 
			[(pocatecni xs)] [(pocatecni xs)]) (prijimaci xs)))) ++
		"\n\nNormovany automat:\n\n" ++
		(aut2str (normovany (del (oznac (vytvoreni (vymaz xs) 
			[(pocatecni xs)] [(pocatecni xs)]) (prijimaci xs))))) ++
		"\nKONEC")

aut2str :: Aut Integer -> String
aut2str [] = ""
aut2str ((xs,ys,zs):ws) = "=== " ++ (sez2str xs) ++ "\b === " ++ 
	(sez2str ys) ++ "\b === " ++ (sez2str zs) ++ "===\n" ++ (aut2str ws)

sez2str :: [Integer] -> String
sez2str [] = ""
sez2str (x:xs) = (show x) ++ " " ++ (sez2str xs)
